
/**
 * @title Praktikum 2: Algorithmendesign (Prof. Ressel)
 * @date 16.05.2013
 *
 */
public class ExpressionTree
{
	private Node root;

	//---START Innere Klasse---//
	public class Node
	{
		public String inhalt;
		public boolean istBlatt;
		Node right, left;
		
		/**
		 * 
		 * @param inhalt
		 * @param istBlatt
		 */
		public Node(String inhalt, boolean istBlatt)
		{
			this.inhalt = inhalt;
			this.istBlatt = istBlatt;
		}
		
		/**
		 * 
		 * @return
		 */
		public boolean istBlatt()
		{
			return istBlatt;
		}
		
		/**
		 * getter Methode fuer Inhalt
		 * @return
		 */
		public String getInhalt()
		{
			return inhalt;
		}
		
		/**
		 * getter Methode fuer Links
		 * @return
		 */
		public Node getLeft()
		{
			return left;
		}
		
		/**
		 * setter Methode fuer Links
		 * @param left
		 */
		public void setLeft(Node left)
		{
			this.left = left;
		}
		
		/**
		 * getter Methode fuer Rechts
		 * @return
		 */
		public Node getRight()
		{
			return right;
		}
		
		/**
		 * setter Methode fuer Rechts
		 * @param right
		 */
		public void setRight(Node right)
		{
			this.right = right;
		}
		
		/**
		 * 
		 * @return
		 */
		String infix()
		{
			if(istBlatt)
			{
				return inhalt;
			}
			else
			{
				return "(" + left.infix() + inhalt + right.infix() + ")";
			}
		}
		
		String postfix()
		{
			if(istBlatt)
			{
				return inhalt;
			}
			else
			{
				return left.postfix() + "" + right.postfix() + "" + inhalt;
			}
		}
		
		String prefix()
		{
			if(istBlatt)
			{
				return inhalt;
			}
			else
			{
				return inhalt + "" + left.prefix() + "" + right.prefix();
			}
		}
		
		/**
		 * 
		 * @return
		 */
		double evaluate()
		{
			if(istBlatt)
			{
				return Double.parseDouble(inhalt);
			}
			else
			{
				inhalt = inhalt.trim();
				double wert = 0;
				char op = inhalt.charAt(0);
				switch (op)
				{
					case '+':
						wert = left.evaluate() + right.evaluate();
						break;
					case '-':
						wert = left.evaluate() - right.evaluate();
						break;
					case '*':
						wert = left.evaluate() * right.evaluate();
						break;
					case '/':
						wert = left.evaluate() / right.evaluate();
						break;
				}
				
				return wert;
			}
		}
		
		public int anzahlNodes()
		{
			if(istBlatt)
			{
				return 1;
			}
			else
			{
				return left.anzahlNodes() + 1 + right.anzahlNodes();
			}
		}
		
		/**
		 * Methode berechnet die Hoehe des Baumes
		 * @return Hoehe des Baumes
		 */
		public int hoehe()
		{
			if(istBlatt)
			{
				return 0;
			}
			else
			{
				return 1 + Math.max(left.hoehe(), right.hoehe());
			}
		}
		
		/**
		 * Methode berechnet die Anzahl der Blaetter des Baumes
		 * @return Anzahl der Blaetter
		 */
		public int anzahlBlaetter()
		{
			if (istBlatt)
			{
				return 1;
			}
			else
			{
				return left.anzahlBlaetter() + right.anzahlBlaetter();
			}
		}
	}
	//---ENDE Innere Klasse---//
	
	public ExpressionTree()
	{
		root = null;
	}
	
	public ExpressionTree(String expression)
	{
		root = insert(expression);
	}
	
	/**
	 * 
	 * @param ex
	 * @return
	 */
	public Node insert(String ex)
	{
		Node asdf;
		ex = ex.trim();
		
		if((ex.charAt(0) == '(') && (ex.charAt(ex.length()-1) == ')'))
		{
			String[] erg = zerlege(ex);
			asdf = new Node(erg[0], false);
			asdf.setLeft(insert(erg[1]));
			asdf.setRight(insert(erg[2]));
		}
		else
		{
			asdf = new Node(ex, true);
		}
		
		return asdf; 
	}
	
	/**
	 * 
	 * @return
	 */
	private String infix()
	{
		if (root != null)
		{
			return root.infix();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	/**
	 * 
	 * @return
	 */
	private String postfix()
	{
		if (root != null)
		{
			return root.postfix();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	/**
	 * 
	 * @return
	 */
	 private String prefix()
	{
		if (root != null)
		{
			return root.prefix();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	/**
	 * 
	 * @return
	 */
	public double evaluate()
	{
		if (root != null)
		{
			return root.evaluate();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	/**
	 * 
	 * @return
	 */
	public int hoehe()
	{
		if (root != null)
		{
			return root.hoehe();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	public int anzahlNodes()
	{
		if (root != null)
		{
			return root.anzahlNodes();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	/**
	 * Gibt die Anzahl der Blaetter aus
	 * @return
	 */
	public double anzahlBlaetter()
	{
		if (root != null)
		{
			return root.anzahlBlaetter();
		}
		else
		{
			throw new RuntimeException("Baum ist leer");
		}
	}
	
	/**
	 * Methode zerlegt den String in drei Teile (rechts, Operator, links)
	 * @param expression den aufzuteilenden String
	 * @return den aufgeteilten String
	 */
	private String[] zerlege(String expression)
	{	
		int counter = 0;
		int pos = 0;
		
		String erg[] = new String[3];
		expression = expression.substring(1, expression.length()-1);

		for(int i = 0; i< expression.length(); i++)
		{   
			char c = expression.charAt(i);
			if(c == '(')
			{
				counter++;
			}
			
			if(c == ')')
			{
				counter--;
			}
			
			if((counter == 0) && ( c == '+' || c == '-' || c == '*' || c == '/'))
			{
				erg[0] = "" + c;
				pos = i;
				break;
			}
		}
		erg[1] = expression.substring(0, pos);
		erg[2] = expression.substring(pos+1, expression.length());
		
		return erg;
	}

	public static void main(String[] args)
	{
		String ex = "(((15-3)*4)+(6/(2+1)))";
		ExpressionTree et = new ExpressionTree(ex);
		System.out.println("Infix-Notation: " + et.infix());
		System.out.println("Postfix-Notation: " + et.postfix());
		System.out.println("Prefix-Notation: " + et.prefix());
		System.out.println("Wert = " + et.evaluate());
		System.out.println("Anzahl Knoten: " + et.anzahlNodes());
		System.out.println("Hoehe = " + et.hoehe());
		System.out.println("Anzahl Blaetter " + et.anzahlBlaetter());
	}
}
